import arcpy
import numpy as np

# Written for ArcGIS 10.1

# The idea here is to estimate a range of "optimal" bandwidths for use in a fixed KDE of a set of points.
# It is assumed that the optimal bandwidths will be used in ArcGIS, hence they are only relevant to the Epanechnikov (Quartic) Kernel.

def getBandwidths(pnts):
    # Could add a range of different Bandwidth techniques here, we'll start with just 4.
    # First 3 are variations of the normal optimal smoothing equation given in Bowman and Azzalini (1997).
    # The adaptation for the Epanechnikov kernel is due to Silverman (1986 p.86)

    # This is the normal optimal smoothing equation: h = 1.77*(sigma*n^-1/6)
    
    nose_h = (0.5 * (np.std(pnts[:,0]) + np.std(pnts[:,1]))) * (float(len(pnts))**(-1.0/6.0)) * 1.77

    # NB as the kernel is symmetrical, sigma is the average spread of x and y.
    
    # Now we can replace the standard deviation, which may lead to oversmoothing of spatial data, with a different measure of spread.

    # This is the Average Absolute Deviation, in which sigma is: mean(abs(xi -mean(x)))

    nose_aad_h = (0.5 * ( np.mean(np.absolute(pnts[:,0]-np.mean(pnts[:,0]))) + np.mean(np.absolute(pnts[:,1]-np.mean(pnts[:,1])))))* (float(len(pnts))**(-1.0/6.0))* 1.77

    # This is the Median Absolute Deviation, in which sigma is: median(abs(xi - median(x)))

    nose_mad_h = (0.5 * ( np.median(np.absolute(pnts[:,0]-np.median(pnts[:,0]))) + np.median(np.absolute(pnts[:,1]-np.median(pnts[:,1])))))* (float(len(pnts))**(-1.0/6.0))* 1.77

    return {"Normal Optimal Smoothing Bandwidth" : nose_h, "Normal Optimal Smoothing Bandwidth; AAD" : nose_aad_h,"Normal Optimal Smoothing Bandwidth; MAD" : nose_mad_h}

# Read the parameter values:
#   1: input point data
#   2: Optional SQL Query
#   3: output table

pointdata = arcpy.GetParameterAsText(0)
sqlQuery = arcpy.GetParameterAsText(1)
OutputTable = arcpy.GetParameterAsText(2)
arcpy.AddMessage(OutputTable)

#pointdata = r"C:\Filestore\OlympicsGIS\FoodPremises\O2\Reclassify\O1_Only_Food_Final_Reclass.shp"
#sqlQuery = "\"NewPrem_O1\" = 15"


# Read in point data

arcpy.MakeFeatureLayer_management(pointdata,"pointdata",sqlQuery)

# Get the row count of the data

rows  = int(arcpy.GetCount_management("pointdata").getOutput(0))

# Create a numpy array to hold the x and y coords.

locs = np.empty((rows,2))

# Now iterate through the points with a cursor and extract the x and y points

rownum = 0

for row in arcpy.da.SearchCursor("pointdata","SHAPE@XY"):
    locs[rownum][0] = float(row[0][0])
    locs[rownum][1] = float(row[0][1])
    rownum += 1

del row

# Now calculate fixed bandwidths

BW_out = getBandwidths(locs)

# Now create an output table for the bandwidths

OutputTable_path = OutputTable[:OutputTable.rfind("\\")]
OutputTable_name = OutputTable[OutputTable.rfind("\\")+1:]

arcpy.CreateTable_management(OutputTable_path,OutputTable_name)
arcpy.MakeTableView_management(OutputTable,"Table")
arcpy.AddField_management("Table","Method","TEXT","","",42)
arcpy.AddField_management("Table","Bandwidth","DOUBLE")
arcpy.DeleteField_management("Table","Field1")

cursor = arcpy.da.InsertCursor("Table",("Method","Bandwidth"))

for i in BW_out.keys():
    cursor.insertRow((i,BW_out[i]))

del cursor


